/******************************************************************************
 *
 *  MiniOS 1.0
 *
 *  Mini multitasking RTOS for microcontrollers
 *  (c) 2005-2010, Hj. Kern
 *
 *  NOTICE:
 *  PLEASE READ THE ASSOCIATED README-FILE.
 *
 *  Project  :   MiniOS
 *  Author   :   Hj. Kern
 *  Date     :   2010
 *  Version  :   1.0
 *  File     :   OS_KERNEL.C
 *
******************************************************************************/

#include "os_minios.h"

// ****************************************************************************
//                         OS CONSTANTS AND VARIABLES
// ****************************************************************************

const OS_TCB_CONST  TaskC[] = { TASK_CONST_DATA };   // TCB const

      OS_TCB        Task[MAX_TASK];                  // TCB vars
      OS_PARAMS     OS;                              // OS parameter

// ****************************************************************************
//                           MICRO KERNEL, SCHEDULER
// ****************************************************************************

// ----------------------------------------------------------------------------
void OS_Init(void)  // init os-params
{
   OS.SysState         = OS_RUN_SYSTEM;  // init os state
   OS.ExceptionCode    = ERR_VOID;    // no error, success
   OS.CurrentTask      = 0;
   OS.JumpMessage      = OSJ_VOID;
   OS.TaskToInit       = MAX_TASK;
   OS.CurrentRRTask    = 0xFF;        // task 0 as first
   OS.MessageBufIndex  = 0;

   OS_ClearTick();
   OS_ClearNotifier();
   OS_ClearMessageBuffer();           // OS.MessageBuf[MAX_MSGBUF+1];
   OS_ClearAllSemaphores();           // OS.Semaphore[MAX_SEM];
   
   OS_InitSem(1,0);
   // init TCB   :  Task[i].Label, State, Timer, Message, Semaphore
   // Label      :  initialized at task function call
   // State      :  "
   // Timer      :  ?
   // Message    :  ?
   // Semaphore  :  ?
}

// ----------------------------------------------------------------------------
void OS_ResumeTask(UBYTE i) // jump to resume task
{
   UBYTE s, msg;
   s = Task[i].State;
   msg = OSJ_VOID;
   if (s == TS_READY_ON_TMR)  msg = OSJ_ON_TMR;   // indication if timeout occured
   OS.CurrentTask = i;
   Task[OS.CurrentTask].State = TS_RUN;
   OS.SysState = OS_RUN_USER;
   longjmp(Task[OS.CurrentTask].Label, msg);
}

// ----------------------------------------------------------------------------
// CHECK EVENT-WAITING TASKS
//
void OS_SetTaskReadyOnEvent(UBYTE tid)
{
   if (Task[tid].State < TS_DELAY) { // no timer involved

      // semaphores signaled?
      if (Task[tid].State == TS_WAIT_SEM) {
         OS_DISABLE_INT
         if (OS.Semaphore[Task[tid].Event]) {     // SEM TEST AND DEC
            OS.Semaphore[Task[tid].Event]--;
            OS_ENABLE_INT
            Task[tid].State = TS_READY_ON_SEM;
            ON_EVENT_DISPATCH_SEM_SIGNALED
            return;
         }
         OS_ENABLE_INT
      }
      // message in buffer?
      else if (OS_GetMessage(Task[tid].Event)) {  // get message or null
         Task[tid].State = TS_READY_ON_MSG;
         ON_EVENT_DISPATCH_MSG_FOUND
         return;
      } // end msg
   }
   else { // timer involved

      // semaphore with timeout?
      if (Task[tid].State == TS_DELAY_SEM) {
         OS_DISABLE_INT
         if (OS.Semaphore[Task[tid].Event]) {     // SEM TEST AND DEC
            OS.Semaphore[Task[tid].Event]--;
            Task[tid].Timer = 0;
            OS_ENABLE_INT
            Task[tid].State = TS_READY_ON_SEM;
            ON_EVENT_DISPATCH_SEM_TIMED
            return;
         }
         OS_ENABLE_INT
      }
      // message with timeout?
      if (Task[tid].State == TS_DELAY_MSG) {
         if (OS_GetMessage(Task[tid].Event)) {    // get message or null
            OS_DISABLE_INT
            Task[tid].Timer = 0;
            OS_ENABLE_INT
            Task[tid].State = TS_READY_ON_MSG;
            ON_EVENT_DISPATCH_MSG_TIMED
            return;
         }
      }
      // task timer expired?
      OS_DISABLE_INT
      if (Task[tid].Timer == 0) {                 // DELAY, timer expired
         OS_ENABLE_INT
         Task[tid].State = TS_READY_ON_TMR;
         ON_EVENT_DISPATCH_TIMER_EXP
         return;
      }
      OS_ENABLE_INT
   } // end checking
} // end fn

// ----------------------------------------------------------------------------
// SCHEDULER, a lowest prio task
//
void OS_Schedule(void)
{
   UBYTE i;
   OS.SysState = OS_RUN_SYSTEM;
   ON_SCHEDULE_ENTRY
   OS_HOOK_SCHEDULE_ENTRY

   if (Task[OS.CurrentTask].State == TS_RUN)     // running tasks get ready
      Task[OS.CurrentTask].State = TS_READY;

   if (OS.TaskToInit) {
      ON_SCHEDULE_TASK_INIT_JUMP
      longjmp(OS.jbExecute, OSJ_VOID);           // no return if sys init
   }

   while (1) {   // schedule loop

      UBYTE bWaitingTaskCount = 0;
      OS_ClearNotifier();   // clear to prevent sys-timeout
      // event-prio-mode if ready-high
      for (i = 0; i < MAX_TASK; i++) {

         // any events for this task?
         if (Task[i].State >= TS_WAIT) {
            OS_SetTaskReadyOnEvent(i);
            if (Task[i].State >= TS_WAIT) bWaitingTaskCount++;   // task still waiting
         }
         // prio-scheduling: first TS_READY_HIGH-task in list has highest prio
         if (Task[i].State >= TS_READY_HIGH && Task[i].State < TS_READY_END) {
            ON_SCHEDULE_EVENT_PRIO_MODE_RESUME
            OS_ResumeTask(i);
         }
      }
      // no event-prio-mode task found until here --> check rr-mode

      // round robin scheduling if TS_READY
      for (i = 0; i < MAX_TASK; i++) { // scan task-array, until next ready-task is encountered
         OS.CurrentRRTask++;
         if (OS.CurrentRRTask == MAX_TASK)  OS.CurrentRRTask = 0;  // clip index at null
         if (Task[OS.CurrentRRTask].State == TS_READY) {
            ON_SCHEDULE_RR_NEW_RESUME
            OS_ResumeTask(OS.CurrentRRTask);
         }
      }
      ON_SCHEDULE_NO_TASK_RESUME

      if (bWaitingTaskCount == 0) {    // no ready AND no waiting tasks in task-list
         OS_Exception(ERR_ALL_TASKS_BLOCKED);   // sys blocked
      }

      OS.SysState = OS_RUN_IDLE;
      ON_SCHEDULE_IDLE_LOOP_ENTRY
      while (OS.NotifyCount == 0) { // waiting tasks, idle-loop while no event,
         OS_HOOK_SCHEDULE_IDLE         // polling jobs, hw-watchdog clear, power management etc.
         ON_SCHEDULE_IDLE_LOOP
      }
      OS.SysState = OS_RUN_SYSTEM;
      ON_SCHEDULE_IDLE_LOOP_EXIT

   } // end schedule loop
} // end schedule

// ----------------------------------------------------------------------------
// os init and execute

void OS_Execute(void) // called in main() at sys-startup, do not recall this function
{
   OS.SysState = OS_STOPPED;   // not yet init'ed
   ON_EXECUTE_ENTRY

   // save context at sys-startup, return null only at sys-startup from main()
   OS.JumpMessage = (UBYTE) setjmp(OS.jbExecute);
   ON_EXECUTE_SETJMP_ENTRY

   // start os (0)
   if (OS.JumpMessage == 0) {
      OS_Init();            // os init all params and set os running
      OS_HOOK_USERINIT_OS   // user init os, enable interrupts
      ON_EXECUTE_INIT_ALL
   }

   // init tasks, schedule
   if (OS.TaskToInit) {
      OS.TaskToInit--;                    // points to next task to init
      OS.CurrentTask = OS.TaskToInit;     // set current task-ID
      OS.JumpMessage = OSJ_VOID;
      Task[OS.CurrentTask].State = TS_RUN;
      OS.SysState = OS_RUN_USER;
      ON_EXECUTE_INITCALL_TASK_FN
      TaskC[OS.CurrentTask].pTaskFn();    // call task-function, no return
      OS_Exception(ERR_TASK_RETURN);      // error if task returns
   }

   // on ISR error, trap: exit via exception --> os_execute
   if (OS.JumpMessage == OSJ_ERROR_ISR_NOTIFY)
      OS_Exception(ERR_ISR_NOTIFY);
   if (OS.JumpMessage == OSJ_ERROR_ISR_SEMOVFL)
      OS_Exception(ERR_ISR_SEMOVFL);

   // terminate os or exit on exception
   ON_EXECUTE_TERMINATE_OS
   OS.SysState = OS_STOPPED;   // exit os
}
// end of OS implementation
// ****************************************************************************