/******************************************************************************
 *
 *  MiniOS 1.0
 *
 *  Mini multitasking RTOS for microcontrollers
 *  (c) 2005-2010, Hj. Kern
 *
 *  NOTICE:
 *  PLEASE READ THE ASSOCIATED README-FILE.
 *
 *  Project  :   MiniOS
 *  Author   :   Hj. Kern
 *  Date     :   2010
 *  Version  :   1.0
 *  File     :   OS_KERNEL.H
 *
******************************************************************************/

#ifndef os_kernel_h
#define os_kernel_h

/******************************************************************************
///////////////////////////////////////////////////////////////////////////////

                            USER CONFIGURATION

                        EDIT THE FOLLOWING SECTION

///////////////////////////////////////////////////////////////////////////////
******************************************************************************/

// ****************************************************************************
//                           USER TASK DATA
// ****************************************************************************

#define MAX_TASK  5   // size of task-array, max 250 tasks

// Task list, insert user task names here, max 250 names
// Task order in list corresponds to task-priority: Task0 has highest Prio.

#define TASK_CONST_DATA    \
                           \
      Task0,               \
      Task1,               \
      Task2,               \
	  Task3,			   \
	  Task4				   \
//    ...

// user task prototypes
void Task0 (void);
void Task1 (void);
void Task2 (void);
void Task3 (void);
void Task4 (void);
// ...

// ****************************************************************************
//                            USER SEMAPHORE DATA
// ****************************************************************************

#define MAX_SEM_COUNT  25  // counting semaphore overflow value: 2...255

// user defined semaphore-array-names/indices 0...250

typedef enum tagUserSemIndex {
   // -------------------------
   // BEGIN OF USER DEFINITIONS
   SEM_00,    // comment
   SEM_01,
   SEM_02,
   SEM_03,
   SEM_04,
   // ...,
   // END OF USER DEFINITIONS
   // -----------------------
   MAX_SEM    // do not edit: max. bytes in semaphore-array
}  OS_USER_SEM_ID;

// USAGE EXAMPLE of function:
//    OS_SignalSem(SEM_00);

// ****************************************************************************
//                       GLOBAL MAILBOX (MESSAGE BUFFER)
// ****************************************************************************

#define MAX_MSGBUF  16   // size of message buffer in bytes, max 250

// user defined event/message names, max 250, can be any binary value 1...250

typedef enum tagOSUserMessage {
   UM_VOID,    // null, NOT USABLE, do not edit
   // -------------------------
   // BEGIN OF USER DEFINITIONS
   MSG_A,      // comment
   MSG_B,
   MSG_C,
   MSG_D,
   // ...,
   // END OF USER DEFINITIONS
   // -----------------------
   MAX_MSG
}  OS_USER_EVENT_MSG;

// USAGE EXAMPLE of function:
//    OS_PostMessage(MSG_A);

// ****************************************************************************
//                        USER ISR, DISABLE, ENABLE
// ****************************************************************************

// define the interrupt disable and enable instructions of your target here

#define OS_DISABLE_INT   cli(); //cli() clears global interrupt flag in SREG
#define OS_ENABLE_INT    sei(); //sei() sets the bit and switches interrupts on

// ****************************************************************************
//                          SYSTEM TIMEOUT CONFIG
// ****************************************************************************

// system timeout, SW-watchdog
// if a task blocks, system terminates on OS_NOTIFY_MAX event counts

#define OS_NOTIFY_MAX  250   // max value 250 (UBYTE)

// ****************************************************************************
//                     USER HOOKS, OS CALLBACK FUNCTIONS
// ****************************************************************************

// user hooks, place callback functions here
#define OS_HOOK_USERINIT_OS       // OnOSUserInit();   // enable interrupts
#define OS_HOOK_EXCEPTION         // OnOSException();
#define OS_HOOK_SCHEDULE_ENTRY    //OnScheduleEntry();   //
#define OS_HOOK_SCHEDULE_IDLE     //OnScheduleIdle();    //

// hook function prototypes
//void OnOSUserInit(void);
//void OnOSException(void);
//void OnScheduleEntry(void);
//void OnScheduleIdle(void);

// ****************************************************************************
//                      USER EXCEPTION/ERROR HANDLING
// ****************************************************************************

// Function OS_Exception(UBYTE Code) can be called to trap errors in the user
// application, insert user error codes here, max 200 code-names:

#define ERR_USER_NAMES             \
                                   \
        ERR_USER_0,                \
        ERR_USER_1,                \
        ERR_USER_2                 \
//      ...

//     USAGE EXAMPLE of function:
//     if (some_user_error)
//        OS_Exception(ERR_USER_0);

// ****************************************************************************
//                         SYSTEM TIMER DATA TYPES
// ****************************************************************************

// if you wish to extend task-timer or tick range: modify this data types

typedef unsigned int     T_TICK;       // sys time, tick counter type
typedef unsigned int     T_TIMER;      // task timer type








/******************************************************************************
///////////////////////////////////////////////////////////////////////////////

                                   SYSTEM

                      DO NOT EDIT THE FOLLOWING SECTION

///////////////////////////////////////////////////////////////////////////////
******************************************************************************/

// ****************************************************************************
//                            SYSTEM DATA TYPES
// ****************************************************************************

typedef unsigned char    BOOL;     // general use
typedef unsigned char    UBYTE;    // general use

// ****************************************************************************
//                           OS DEFINES, MACROS
// ****************************************************************************

#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE !FALSE
#endif
#ifndef NULL
#define NULL 0
#endif
#ifndef SUCCESS
#define SUCCESS 1
#endif

// debug hooks
#ifndef DEBUG_ON
// api
// os control
#define ON_OS_EXCEPTION
#define ON_OS_TERMINATE
// task control
#define ON_OS_STOP_CURRENT_TASK
#define ON_OS_STOP_TASK
#define ON_OS_START_TASK
#define ON_OS_START_TASK_ISR
#define ON_OS_START_BACKGROUND_TASK
// timer
#define ON_OS_TIMERTICK
#define ON_OS_TIMERTICK_TIMER_EXP
#define ON_OS_SET_TIMER
#define ON_OS_WAIT_TIMER
// semaphore
#define ON_OS_SIGNAL_SEM
#define ON_OS_SIGNAL_ISR_SEM
#define ON_OS_WAIT_SEM
// message box
#define ON_OS_PEEK_MESSAGE
#define ON_OS_WAIT_MESSAGE
#define ON_OS_POST_MESSAGE_SUCCESS
#define ON_OS_GET_BUFMESSAGE_SUCCESS
#define ON_OS_GET_BUFMESSAGE_EMPTY
#define ON_OS_GET_BUFMESSAGE_NOMSG
// kernel
// set task ready on event (event dispatch)
#define ON_EVENT_DISPATCH_SEM_SIGNALED
#define ON_EVENT_DISPATCH_MSG_FOUND
#define ON_EVENT_DISPATCH_SEM_TIMED
#define ON_EVENT_DISPATCH_MSG_TIMED
#define ON_EVENT_DISPATCH_TIMER_EXP
// schedule
#define ON_SCHEDULE_ENTRY
#define ON_SCHEDULE_TASK_INIT_JUMP
#define ON_SCHEDULE_EVENT_PRIO_MODE_RESUME
#define ON_SCHEDULE_RR_NEW_RESUME
#define ON_SCHEDULE_NO_TASK_RESUME
#define ON_SCHEDULE_IDLE_LOOP_ENTRY
#define ON_SCHEDULE_IDLE_LOOP
#define ON_SCHEDULE_IDLE_LOOP_EXIT
// os execute
#define ON_EXECUTE_ENTRY
#define ON_EXECUTE_SETJMP_ENTRY
#define ON_EXECUTE_INIT_ALL
#define ON_EXECUTE_INITCALL_TASK_FN
#define ON_EXECUTE_TERMINATE_OS
//
#endif // DEBUG_HOOKS

// ****************************************************************************
//                       OS TYPES, CONSTANTS, VARIABLES
// ****************************************************************************

typedef enum tagSysState {   // OS.SysState variable
   OS_STOPPED,
   OS_RUNNING,       // generic running state
   OS_RUN_SYSTEM,    // running kernel/scheduler
   OS_RUN_IDLE,      // running idle loop
   OS_RUN_USER,      // running task
   //
   OS_USER_STATES,   // user defined states
   //
   MAX_SYS_STATE
}  SYS_STATE;

typedef enum tagOSError {
   ERR_VOID,                 // no error, success
   ERR_EXIT_OS,              // undefined exit (unused in this version)
   ERR_ARRAY_BOUNDARY,       // array boundary, array size error
   ERR_TASK_RETURN,          // task function returns
   ERR_EVENT_NOTIFY,         // event counter ovfl, task blocked
   ERR_ALL_TASKS_BLOCKED,    // no ready, no waiting tasks
   ERR_SEM_OVFL,             // sem ovfl
   ERR_MSGBUF_OVFL,          // msgbuf full
   //
   ERR_ISR_NOTIFY,           // ISR: sys-timeout, task blocked
   ERR_ISR_SEMOVFL,          // ISR: sem ovl
   //
   ERR_USER_NAMES,           // user defined errors
   //
   MAX_ERR
}  OS_ERROR;

typedef enum tagOSTaskState {
   TS_VOID,              // no effect
   TS_RUN,
   TS_STOP,
   TS_TERMINATE,         // not used in this version
   // ready-states
   TS_READY,             // ready, without event, rr-mode
   TS_READY_HIGH,        // ready on event, prio-scheduling
   TS_READY_ON_TMR,      // timer event
   TS_READY_ON_SEM,      // semaphore event
   TS_READY_ON_MSG,      // message event
   TS_READY_END,         // eof ready-states
   // wait-states, without timer
   TS_WAIT,              // task waiting
   TS_WAIT_SEM,          // for semaphore
   TS_WAIT_MSG,          // for message (state not used in this version)
   //w timer
   TS_DELAY,             // task timer
   TS_DELAY_SEM,         // sem w timeout
   TS_DELAY_MSG,         // msg w timeout
   //
   TS_WAIT_END,          // eof wait states
   //
   MAX_TS
}  OS_TSTATE;

typedef enum tagOSJumpMsg {   // longjmp() return values
   OSJ_SYS_STARTUP,           // null: direct from setjmp()
   OSJ_VOID,                  // no effect
   OSJ_ERROR_ISR_NOTIFY,      // state info for os_exec
   OSJ_ERROR_ISR_SEMOVFL,
   OSJ_ON_TMR,                // state info for task: task timer exp
   //
   MAX_OSJ
}  OS_JUMP_MSG;

typedef struct {      // TCB const in code memory
   void (* pTaskFn)(void);      // task function pointer
}  OS_TCB_CONST;

typedef struct {      // TCB in data memory, GLOBAL
   jmp_buf       Label;         // os task context
   UBYTE         State;         // task[i] state
   T_TIMER       Timer;         // task timer counted down if delay-state
   UBYTE         Event;         // message or semaphore task is waiting for
}  OS_TCB;

extern OS_TCB Task[];

typedef struct {      // os params, GLOBAL
   UBYTE     SysState;           // os state, os terminated == 0, else >0
   jmp_buf   jbExecute;          // os start state/label
   T_TICK    TickCount;          // sys time
   UBYTE     NotifyCount;        // counts incoming events
   UBYTE     ExceptionCode;      // os exception
   UBYTE     CurrentTask;        // running task
   UBYTE     JumpMessage;        // task switch: longjmp() return value
   UBYTE     TaskToInit;         // index to next init-task, null if no init
   UBYTE     CurrentRRTask;      // indicates task that is in rr-mode
   UBYTE     Semaphore[MAX_SEM]; // semaphore array, indexed by Task[i].Event
   UBYTE     MessageBuf[MAX_MSGBUF+1]; // task message buf plus null-termination as string
   UBYTE     MessageBufIndex;    // next free item location in msg-buf, shows buf-item count
}  OS_PARAMS;

extern OS_PARAMS OS;

// ****************************************************************************
//                            OS EXPORT FUNCTIONS
// ****************************************************************************

void OS_Init(void);
void OS_Schedule(void);
void OS_Execute(void);

#endif // end of header
// ****************************************************************************