/******************************************************************************
 *
 *  MiniOS 1.0
 *
 *  Mini multitasking RTOS for microcontrollers
 *  (c) 2005-2010, Hj. Kern
 *
 *  NOTICE:
 *  PLEASE READ THE ASSOCIATED README-FILE.
 *
 *  Project  :   MiniOS
 *  Author   :   Hj. Kern
 *  Date     :   2010
 *  Version  :   1.0
 *  File     :   OS_USERAPI.C
 *
******************************************************************************/

#include "os_minios.h"

// ****************************************************************************
//                              OS CONTROL
// ****************************************************************************

// ----------------------------------------------------------------------------
void OS_Exception(UBYTE Code)
{
   OS.ExceptionCode = Code;
   OS_HOOK_EXCEPTION
   ON_OS_EXCEPTION
   longjmp(OS.jbExecute, OSJ_VOID);   // trap: terminate os
}

// ----------------------------------------------------------------------------
void OS_ClearException(void)
{
   OS.ExceptionCode  = ERR_VOID;
}

// ----------------------------------------------------------------------------
// counts the number of incoming events, non ISR
//
void OS_NotifyEvent(void)
{
   OS_DISABLE_INT
   OS.NotifyCount++;                        // cleared again in scheduler
   if (OS.NotifyCount >= OS_NOTIFY_MAX) {
      OS_ENABLE_INT
      OS_Exception(ERR_EVENT_NOTIFY);       // event overflow
   }
   OS_ENABLE_INT
}

// ----------------------------------------------------------------------------
// **** ISR level ONLY
// sys timeout, counts sys-timer-ISR, terminates the system if blocked
// if error: a task is probably blocking the system
//
void OS_Notify_ISR(void)
{
   OS_DISABLE_INT
   OS.NotifyCount++;                        // cleared again in scheduler

   if (OS.NotifyCount >= OS_NOTIFY_MAX) {   // test if ISR event overflow
      if (OS.SysState != OS_STOPPED) {      // test if os terminated
         OS_ENABLE_INT
         longjmp(OS.jbExecute, OSJ_ERROR_ISR_NOTIFY);
      }
      else {                       // if os terminated: no ISR notification
         OS.NotifyCount = OS_NOTIFY_MAX;
      }
   }
   OS_ENABLE_INT
}

// ----------------------------------------------------------------------------
// **** ISR and TASK level
void OS_ClearNotifier(void)   // to prevent sys timeout in user app
{
   OS_DISABLE_INT
   OS.NotifyCount = 0;
   OS_ENABLE_INT
}

// ----------------------------------------------------------------------------
// **** ISR and TASK level
UBYTE OS_PeekNotifier(void)
{
   UBYTE i;
   OS_DISABLE_INT
   i = OS.NotifyCount;
   OS_ENABLE_INT
   return i;
}

// ----------------------------------------------------------------------------
// **** ISR and TASK level
void OS_Terminate(void)
{
   ON_OS_TERMINATE
   longjmp(OS.jbExecute, OSJ_VOID);
}

// ****************************************************************************
//                             OS TASK CONTROL
// ****************************************************************************

// ----------------------------------------------------------------------------
void OS_StopCurrentTask(void)
{
   Task[OS.CurrentTask].State = TS_STOP;
   ON_OS_STOP_CURRENT_TASK
}

// ----------------------------------------------------------------------------
void OS_StopTask(UBYTE tid)
{
   if (tid >= MAX_TASK) OS_Exception(ERR_ARRAY_BOUNDARY);
   Task[tid].State = TS_STOP;
   ON_OS_STOP_TASK
}

// ----------------------------------------------------------------------------
void OS_StartTask(UBYTE tid)
{
   if (tid >= MAX_TASK) OS_Exception(ERR_ARRAY_BOUNDARY);
   Task[tid].State = TS_READY_HIGH;
   OS_NotifyEvent();
   ON_OS_START_TASK
}

// ----------------------------------------------------------------------------
void OS_StartBackgroundTask(UBYTE tid)
{
   if (tid >= MAX_TASK) OS_Exception(ERR_ARRAY_BOUNDARY);
   Task[tid].State = TS_READY;
   OS_NotifyEvent();
   ON_OS_START_BACKGROUND_TASK
}

// ****************************************************************************
//                                 OS TIMER
// ****************************************************************************

// ----------------------------------------------------------------------------
// **** ISR Timer level, counts sys-timer ticks, services task-timer
// must be called inside a user timer ISR
void OS_TimerTick(void)
{
   UBYTE i;

   OS.TickCount++;   // sys time
   ON_OS_TIMERTICK

   for (i = 0; i < MAX_TASK; i++) {   // task timer decrement
      if (Task[i].Timer > 0) {
         Task[i].Timer--;
         if (Task[i].Timer == 0) {
            ON_OS_TIMERTICK_TIMER_EXP
            //OS_Notify_ISR();
         }
      }
   }
   OS_Notify_ISR();
}

// ----------------------------------------------------------------------------
// **** ISR and TASK level
void OS_ClearTick(void)
{
   OS_DISABLE_INT
   OS.TickCount = 0;
   OS_ENABLE_INT
}

// ----------------------------------------------------------------------------
// **** ISR and TASK level
void OS_SetTick(T_TICK Ticks)
{
   OS_DISABLE_INT
   OS.TickCount = Ticks;
   OS_ENABLE_INT
}

// ----------------------------------------------------------------------------
// **** ISR and TASK level
T_TICK OS_PeekTick(void)
{
   T_TICK i;
   OS_DISABLE_INT
   i = OS.TickCount;
   OS_ENABLE_INT
   return i;
}

// ----------------------------------------------------------------------------
void OS_ClearTimer(void) // reset task timer, no waiting
{
   OS_DISABLE_INT
   Task[OS.CurrentTask].Timer = 0;
   OS_ENABLE_INT
}

// ----------------------------------------------------------------------------
void OS_SetTimer(T_TIMER Ticks) // set and start task timer, no waiting
{
   OS_DISABLE_INT
   Task[OS.CurrentTask].Timer = Ticks;
   OS_ENABLE_INT
   ON_OS_SET_TIMER
}

// ----------------------------------------------------------------------------
void OS_WaitTimer(T_TIMER Ticks) // set and wait task timer
{
   Task[OS.CurrentTask].State = TS_DELAY;
   OS_DISABLE_INT
   Task[OS.CurrentTask].Timer = Ticks;
   OS_ENABLE_INT
   ON_OS_WAIT_TIMER
}

// ----------------------------------------------------------------------------
T_TIMER OS_PeekTimer(void)
{
   T_TIMER i;
   OS_DISABLE_INT
   i = Task[OS.CurrentTask].Timer;
   OS_ENABLE_INT
   return i;
}

// ----------------------------------------------------------------------------
// BLOCKING HALT: stops execution of tasks and task switching
// can also be used outside of OS_Execute() if timer interrupt is enabled
void OS_Halt(T_TIMER Ticks)
{
   OS_SetTimer(Ticks);
   while (OS_PeekTimer()) OS_ClearNotifier();
}

// ****************************************************************************
//                              OS SEMAPHORES
// ****************************************************************************

// ----------------------------------------------------------------------------
// **** ISR and TASK level
void OS_InitSem(UBYTE SemID, UBYTE n)
{
   if (SemID >= MAX_SEM) OS_Exception(ERR_ARRAY_BOUNDARY);

   if (n >= (MAX_SEM_COUNT - 1)) {   // limit init-value
      n = (MAX_SEM_COUNT - 1);
   }
   OS_DISABLE_INT
   OS.Semaphore[SemID] = n;
   OS_ENABLE_INT
}

// ----------------------------------------------------------------------------
// **** ISR and TASK level
void OS_ClearSem(UBYTE SemID)
{
   if (SemID >= MAX_SEM) OS_Exception(ERR_ARRAY_BOUNDARY);
   OS_DISABLE_INT
   OS.Semaphore[SemID] = 0;
   OS_ENABLE_INT
}

// ----------------------------------------------------------------------------
void OS_ClearAllSemaphores(void)
{
   UBYTE i;
   OS_DISABLE_INT
   for (i = 0; i < MAX_SEM; i++) {
      OS.Semaphore[i] = 0;
   }
   OS_ENABLE_INT
}

// ----------------------------------------------------------------------------
// **** ISR and TASK level
// TEST IF SEMAPHORE OVERFLOWED
// in:   sem index of tested sem
// out:  null if sem overflowed, else: number of counts left
//
UBYTE OS_TestSemOvfl(UBYTE SemID)
{
   UBYTE i;
   if (SemID >= MAX_SEM) OS_Exception(ERR_ARRAY_BOUNDARY);
   OS_DISABLE_INT
   i = OS.Semaphore[SemID];
   OS_ENABLE_INT
   return ((MAX_SEM_COUNT - 1) - i); // return null if ovfl
}

// ----------------------------------------------------------------------------
// signal semaphore in task, non blocking
void OS_SignalSem(UBYTE SemID)
{
   if (SemID >= MAX_SEM) OS_Exception(ERR_ARRAY_BOUNDARY);

   OS_DISABLE_INT
   OS.Semaphore[SemID]++;
   if (OS.Semaphore[SemID] >= MAX_SEM_COUNT) {
      OS_ENABLE_INT
      OS_Exception(ERR_SEM_OVFL); // trap exception
   }
   OS_ENABLE_INT
   ON_OS_SIGNAL_SEM
   OS_NotifyEvent();
}

// ----------------------------------------------------------------------------
// **** ISR level ONLY
// signal semaphore in ISR, non blcking
void OS_SignalSem_ISR(UBYTE SemID)          // signal semaphore in ISR
{
   if (SemID >= MAX_SEM) OS_Exception(ERR_ARRAY_BOUNDARY);

   OS_DISABLE_INT
   OS.Semaphore[SemID]++;

   if (OS.Semaphore[SemID] >= MAX_SEM_COUNT) {
      if (OS.SysState != OS_STOPPED) {      // test if os terminated
         OS_ENABLE_INT
         longjmp(OS.jbExecute, OSJ_ERROR_ISR_SEMOVFL);   // trap exception
      }
      else {                       // if os terminated: no ISR-Sem notification
         OS.Semaphore[SemID] = MAX_SEM_COUNT;
      }
   }
   OS_ENABLE_INT
   ON_OS_SIGNAL_ISR_SEM
   OS_Notify_ISR();
}

// ----------------------------------------------------------------------------
void OS_WaitSem(UBYTE SemID, T_TIMER Ticks)
{
   if (SemID >= MAX_SEM) OS_Exception(ERR_ARRAY_BOUNDARY);

   if (Ticks) {
      Task[OS.CurrentTask].State  = TS_DELAY_SEM; // with timeout
      OS_DISABLE_INT
      Task[OS.CurrentTask].Timer = Ticks;
      OS_ENABLE_INT
   }
   else Task[OS.CurrentTask].State = TS_WAIT_SEM; // no timeout
   Task[OS.CurrentTask].Event = SemID;
   ON_OS_WAIT_SEM
}

// ----------------------------------------------------------------------------
// **** ISR and TASK level
UBYTE OS_PeekSem(UBYTE SemID)
{
   UBYTE i;
   if (SemID >= MAX_SEM) OS_Exception(ERR_ARRAY_BOUNDARY);
   OS_DISABLE_INT
   i = OS.Semaphore[SemID];
   OS_ENABLE_INT
   return i;
}

// ****************************************************************************
//                         OS MESSAGE BUFFER, MAILBOX
// ****************************************************************************
//
// item: message-byte
// MAX_MSGBUF: max number of items, index: buf[0 ... MAX_MSGBUF-1]
//
// OS.MessageBufIndex indexes next free item location
//                                                index MAX_MSGBUF
//                                                         |
//  bottom of buffer     top-item          top of buffer   |
//         |                 |                       |
//      |-----|-----|-----|-----|-----|-----|-----|-----|-----|
//  Msg   ME4   ME1   ME3   ME7    x     x     x     x   '\0'   buf as string
//      |-----|-----|-----|-----|-----|-----|-----|-----|-----|
//  idx    0     1     2     3     4     5     6     7     8
//         |                       |
//      empty <--> fill    OS.MessageBufIndex
//
//  This is a stack-like message-buffer: index OS.MessageBufIndex points to next
//  free input location, where incoming messages are stored on the top of the buffer.
//  Messages in the buffer are not held in a chronological order,
//  they get mixed up while searching the buffer for a specific message to output.

// ----------------------------------------------------------------------------
// CLEAR ALL MESSAGES OF THE SAME TYPE IN BUFFER, non blocking
//
// in:   message to clear
// out:  count of specific message if found and cleared, else null
//
UBYTE OS_ClearMessage(UBYTE Msg)
{
   UBYTE EqualCount = 0;
   UBYTE i = 0;

   if (OS.MessageBufIndex == 0) return 0; // return null if buffer empty

   // scan buffer for given message
   for (i = OS.MessageBufIndex - 1; i < OS.MessageBufIndex; i--) {
      if (OS.MessageBuf[i] == Msg) {
         EqualCount++;
         // found: fill the gap with top-most-item in buffer
         OS.MessageBuf[i] = OS.MessageBuf[OS.MessageBufIndex - 1];
         OS.MessageBuf[OS.MessageBufIndex - 1] = 'x';   // debug fill
         OS.MessageBufIndex--;                          // and put the index to item below
      }
   }
   return EqualCount; // null if not found this message item
}

// ----------------------------------------------------------------------------
// CLEAR ALL MESSAGES IN BUFFER
//
void OS_ClearMessageBuffer(void)
{
   UBYTE i = 0;

   // debug fill buf and make a string out of buffer content
   for (i = 0; i < MAX_MSGBUF; i++)  OS.MessageBuf[i] = 'x';

   OS.MessageBuf[MAX_MSGBUF] = '\0';
   OS.MessageBufIndex = 0;
}

// ----------------------------------------------------------------------------
// GET NUMBER OF MESSAGES IN BUFFER
//
// out:  current number of available byte-messages in buffer
//
UBYTE OS_GetMessageCount(void)
{
   return OS.MessageBufIndex;
}

// ----------------------------------------------------------------------------
// GET FREE MAILBOX SPACE, test if mailbox overflow
//
// out:  null if no more space, else: number of locations left
//
UBYTE OS_GetMailboxSpace(void)
{
   // if OS.MessageBufIndex == MAX_MSGBUF --> buffer full
   return (MAX_MSGBUF - OS.MessageBufIndex); // return null if no more space
}

// ----------------------------------------------------------------------------
// WRITE MESSAGE TO BUFFER, non blocking
// no return, if buf full --> exception
//
void OS_PostMessage(UBYTE Msg)
{
   if (OS.MessageBufIndex == MAX_MSGBUF) {    // check if buffer full
      OS_Exception(ERR_MSGBUF_OVFL);
   }
   // write buf
   OS.MessageBuf[OS.MessageBufIndex] = Msg;   // store message
   OS.MessageBufIndex++;                      // index next item
   ON_OS_POST_MESSAGE_SUCCESS
   OS_NotifyEvent();
}

// ----------------------------------------------------------------------------
// WAIT FOR SPECIFIC MESSAGE AND REMOVE IT FROM BUFFER, with timeout
//
void OS_WaitMessage(UBYTE Msg, T_TIMER Ticks)
{
   if (Ticks) {
      Task[OS.CurrentTask].State  = TS_DELAY_MSG; // with timeout
      OS_DISABLE_INT
      Task[OS.CurrentTask].Timer = Ticks;
      OS_ENABLE_INT
   }
   else Task[OS.CurrentTask].State = TS_WAIT;     // no timeout
   Task[OS.CurrentTask].Event = Msg;
   ON_OS_WAIT_MESSAGE
}

// ----------------------------------------------------------------------------
// READ AND REMOVE specific message item from buffer if it exists,
// non blocking
//
// in:   message searched for
// out:  null if buf empty or message not found, else: searched message
//
UBYTE OS_GetMessage(UBYTE Msg)
{
   UBYTE i = 0;
   if (OS.MessageBufIndex == 0) { // return null if buffer empty
      ON_OS_GET_BUFMESSAGE_EMPTY
      return 0;
   }
   // scan buffer for given message
   for (i = OS.MessageBufIndex - 1; i < OS.MessageBufIndex; i--) {
      if (OS.MessageBuf[i] == Msg) {
         // found: fill the gap with top-most-item in buffer
         OS.MessageBuf[i] = OS.MessageBuf[OS.MessageBufIndex - 1];
         OS.MessageBuf[OS.MessageBufIndex - 1] = 'x';   // debug fill
         OS.MessageBufIndex--;                          // and put the index to item below
         ON_OS_GET_BUFMESSAGE_SUCCESS
         return Msg;
      }
   }
   ON_OS_GET_BUFMESSAGE_NOMSG
   return 0;   // not found this message item
}

// ----------------------------------------------------------------------------
// CHECK IF SPECIFIC MESSAGE IN BUFFER, no change of buffer, non blocking
//
// in:   message searched for
// out:  count of specific message if found, else null
//
UBYTE OS_PeekMessage(UBYTE Msg)
{
   UBYTE EqualCount = 0;
   UBYTE i = 0;

   if (OS.MessageBufIndex == 0) return 0; // return null if buffer empty

   // scan buffer for given message
   for (i = OS.MessageBufIndex - 1; i < OS.MessageBufIndex; i--) {
      if (OS.MessageBuf[i] == Msg) EqualCount++;
   }
   ON_OS_PEEK_MESSAGE
   return EqualCount;
}

// end api
// ****************************************************************************