/******************************************************************************
 *
 *  MiniOS 1.0
 *
 *  Mini multitasking RTOS for microcontrollers
 *  (c) 2005-2010, Hj. Kern
 *
 *  NOTICE:
 *  PLEASE READ THE ASSOCIATED README-FILE.
 *
 *  Project  :   MiniOS
 *  Author   :   Hj. Kern
 *  Date     :   2010
 *  Version  :   1.0
 *  File     :   OS_USERAPI.H
 *
******************************************************************************/

#ifndef os_userapi_h
#define os_userapi_h

// macros, user shortcuts

#define STATE        OS.SysState
#define CURRENT      OS.CurrentTask
#define ERROR        OS.ExceptionCode

#define CLR_NOTIFY() OS_ClearNotifier();  // event notifier clr
#define NOTIFY       OS_PeekNotifier()    // returns UBYTE of event notifier

#define CLR_TICK()   OS_ClearTick();      // system tick clr
#define TICK         OS_PeekTick()        // returns T_TICK of sys tick counter

#define CLR_TIMER()  OS_ClearTimer();     // current task timer clr
#define TIMER        OS_PeekTimer()       // returns T_TIMER val of current task timer

#define ON_TIMEOUT  (OS.JumpMessage == OSJ_ON_TMR) // test if task returns on timeout

// save task context and schedule
#define SCHEDULE()                                                         \
         {                                                                 \
            OS.JumpMessage = (UBYTE) setjmp(Task[OS.CurrentTask].Label);   \
            if (OS.JumpMessage == 0) OS_Schedule();                        \
         }
//

#define SCHED_PRIO()                                                       \
        {                                                                  \
           OS.JumpMessage = (UBYTE) setjmp(Task[OS.CurrentTask].Label);    \
           if (OS.JumpMessage == 0) {                                      \
              if (Task[OS.CurrentTask].State == TS_RUN)                    \
                 Task[OS.CurrentTask].State = TS_READY_HIGH;               \
              OS_Schedule();                                               \
           }                                                               \
        }
//

// pass to other task:
//      set other task ready_high (prio-mode)
//      stop current task, save context and schedule
#define STOP_START(tid)  \
        {                                                                  \
           Task[tid].State = TS_READY_HIGH;                                \
           OS_NotifyEvent();                                               \
           Task[OS.CurrentTask].State = TS_STOP;                           \
           OS.JumpMessage = (UBYTE) setjmp(Task[OS.CurrentTask].Label);    \
           if (OS.JumpMessage == 0) OS_Schedule();                         \
        }
//

#define STOP()                                                             \
         {                                                                 \
            Task[OS.CurrentTask].State = TS_STOP;                          \
            OS.JumpMessage = (UBYTE) setjmp(Task[OS.CurrentTask].Label);   \
            if (OS.JumpMessage == 0) OS_Schedule();                        \
         }
//

#define WAIT(Ticks)                                                        \
         {                                                                 \
            Task[OS.CurrentTask].State = TS_DELAY;                         \
            OS_DISABLE_INT                                                 \
            Task[OS.CurrentTask].Timer = Ticks;                            \
            OS_ENABLE_INT                                                  \
            OS.JumpMessage = (UBYTE) setjmp(Task[OS.CurrentTask].Label);   \
            if (OS.JumpMessage == 0) OS_Schedule();                        \
         }
//

#define WAIT_SEM(SemID)                                                    \
         {                                                                 \
            Task[OS.CurrentTask].State = TS_WAIT_SEM;                      \
            Task[OS.CurrentTask].Event = SemID;                            \
            OS.JumpMessage = (UBYTE) setjmp(Task[OS.CurrentTask].Label);   \
            if (OS.JumpMessage == 0) OS_Schedule();                        \
         }
//

#define WAIT_MSG(Msg)                                                      \
         {                                                                 \
            Task[OS.CurrentTask].State = TS_WAIT;                          \
            Task[OS.CurrentTask].Event = Msg;                              \
            OS.JumpMessage = (UBYTE) setjmp(Task[OS.CurrentTask].Label);   \
            if (OS.JumpMessage == 0) OS_Schedule();                        \
         }
//

// function-prototypes of os_userapi.c

// os control
void OS_Exception(UBYTE Code);
void OS_ClearException(void);
void OS_NotifyEvent(void);
void OS_Notify_ISR(void);
void OS_ClearNotifier(void);
UBYTE OS_PeekNotifier(void);
void OS_Terminate(void);
// os task control
void OS_StopCurrentTask(void);
void OS_StopTask(UBYTE tid);
void OS_StartTask(UBYTE tid);
void OS_StartBackgroundTask(UBYTE tid);
// os timer
void OS_TimerTick(void);
void OS_ClearTick(void);
void OS_SetTick(T_TICK Ticks);
T_TICK OS_PeekTick(void);
void OS_ClearTimer(void);
void OS_SetTimer(T_TIMER Ticks);
void OS_WaitTimer(T_TIMER Ticks);
T_TIMER OS_PeekTimer(void);
void OS_Halt(T_TIMER Ticks);
// os semaphores
void OS_InitSem(UBYTE SemID, UBYTE n);
void OS_ClearSem(UBYTE SemID);
void OS_ClearAllSemaphores(void);
UBYTE OS_TestSemOvfl(UBYTE SemID);
void OS_SignalSem(UBYTE SemID);
void OS_SignalSem_ISR(UBYTE SemID);
void OS_WaitSem(UBYTE SemID, T_TIMER Ticks);
UBYTE OS_PeekSem(UBYTE SemID);
// os message buffer, mailbox
UBYTE OS_ClearMessage(UBYTE Msg);
void OS_ClearMessageBuffer(void);
UBYTE OS_GetMessageCount(void);
UBYTE OS_GetMailboxSpace(void);
void OS_PostMessage(UBYTE Msg);
void OS_WaitMessage(UBYTE Msg, T_TIMER Ticks);
UBYTE OS_GetMessage(UBYTE Msg);
UBYTE OS_PeekMessage(UBYTE Msg);

#endif // end of header
// ****************************************************************************